/************************************************************************
 Author: Umit Ogras  (uogras@ece.cmu.edu)
 File name: SmallNoC.cpp
 Date Written:  <Mon Feb 28 19:16:26 2005>
 Last Modified:
 Description: Add long link to mesh network based on input traffic
 Usage:
  ./SmallNoC -number_of_rows 4 -number_of_columns 4 -constraint 10 
-traffic_file traffic.config -architecture_file output.config
************************************************************************/

#include <iostream>
#include <fstream>
#include <cstdio>
#include <vector>

#define MAX_LINE 1024
using namespace std;

int gen_arch_config(int rows,int cols,int count,vector<int> long_links,char *arch_file);
double free_packet_delay(int row, int col, char *traf_config, char *arch_config);
void print_help(void);
bool parse_parameters(int argn, char ** argv,int &row, int &col, int &constraint, 
                      char **arch_file,char **traffic_file, bool &verbose);

int main(int argn, char **argv) {
 
  int num_of_rows=4,num_of_columns=4;                  // Dimensions of the underlying mesh network
  int constraint=10;                                   // Constraint on the number of additional
                                                       // long-range links
  char * arch_file=0, * traffic_file=0;                // Pointers to the traffic and architecture files
  bool verbose = 0;                                    // In the verbose mode print out details 
  
  // Parse the command line parameters
  if (argn == 1) {
    print_help();
    return 0;
  }   
  
  if (!parse_parameters(argn,argv,num_of_rows,num_of_columns,constraint,
                         &arch_file,&traffic_file,verbose) ) {
      cerr<< "Error parsing parameters \n";
      return 0;
    }
  
  if (verbose) {
    cout << "Running the long-range link insertion tool with parameters:\n"
         << "number of rows:\t" << num_of_rows <<endl
         << "number of columns:\t" << num_of_columns <<endl
         << "constraint:\t" << constraint <<endl
         << "traffic file:\t" << traffic_file <<endl
         << "architecture file: " << arch_file <<endl;
  }

  int N = num_of_rows*num_of_columns;                                      // Network size
  
  // Main loop: Continue until the utilization is less than the constraint.
  // Each time find the most beneficial link to be added
  
  /*** Initialization before the loop starts ***/
  bool link_add [N];                        // Link_add[i]=1 => A long_range link is attached to node i
                                            // each node can have at most one long link 
  for (unsigned int r=0; r<N; r++)
    link_add[r]=0;                          // No link is added initially
    
  vector <int> long_links;                  // Long_range links inserted to the network
  
  // To generate the initial topology and routing table (which is mesh and XY routing),
  // we do a trick: A link is added between nodes 0 and 1, after generatinng the architecture
  // files, we delete the link.
  long_links.push_back(0);
  long_links.push_back(1);
  unsigned int count = 2;                    // Number of long links already added
  gen_arch_config(num_of_rows,num_of_columns,count,long_links,arch_file);  
  double ori_tao = free_packet_delay(num_of_rows,num_of_columns,traffic_file,arch_file); 
  long_links.pop_back();
  long_links.pop_back();

  int utilization=0;                         // Total length of long_range links  inserted sofar
  double min_rate=0;                        
  double curr_rate=0;                        // Free packet delay of the current configuration
  int best_id[2]={0,0};                      // Best set of links obtained in the loop
  
  /*** End of the initialization before the main loop  ***/
  //Starting the main loop  
  unsigned int iteration = 0;
  while ( utilization < constraint ) {
    min_rate = 1000;
    // Try all possible additions of links
    for (unsigned int m=0; m<N; m++) {
      for (int n=0; n<N; n++) {
        if ( link_add[m] || link_add[n] || (m==n) )
          continue;                                            // A link already exists between m and n
     
        // Generate config file with this link added
        long_links.push_back(m);
        long_links.push_back(n);
        
        // Write architecture config file with routing info to the architecture_file
        gen_arch_config(num_of_rows,num_of_columns,count,long_links,arch_file);
        
        // Evaluate current configuration
        curr_rate = free_packet_delay(num_of_rows,num_of_columns,traffic_file,arch_file); 
               
        if ( curr_rate < min_rate ) {                          // A better configuration is found
          best_id[0] = m;
          best_id[1] = n;
          min_rate = curr_rate;
          if (verbose)
            printf("Iteration %d, Found link: (%d,%d), Improved free packet delay:%f\n",iteration,m,n,curr_rate);
        } // if
        
        long_links.pop_back();                     // Remove the long_range link for next iteration
        long_links.pop_back();
      } // for c
    } // for r
    
    long_links.push_back(best_id[0]);              // Insert the best link found in the loop permanently
    long_links.push_back(best_id[1]);
    link_add[best_id[0]] = 1;
    link_add[best_id[1]] = 1;
    count += 2;
    
    // Update the utilization
    int dy = abs( (best_id[0] % num_of_columns)-(best_id[1] % num_of_columns));
    int dx =abs( ((int) (best_id[0]/num_of_rows))-((int) (best_id[1]/num_of_rows)) );
    utilization = utilization + dy + dx;

    if (verbose)
      printf("Iteration %d, Permanently inserted the long_range link (%d,%d)\n",
             iteration,best_id[0],best_id[1]);
    
    iteration++;
  } // End of main loop 
  
  // Print out the long_range links inserted to the original network
  cout <<"\n**********************************************************\n\n";
  cout <<"The following long range links are inserted to the network:\n";
  for (int h=0;h<long_links.size();h++)
    printf("%d ",long_links[h]);
  cout << endl;
 
  gen_arch_config(num_of_rows,num_of_columns,count-2,long_links,arch_file);
  curr_rate = free_packet_delay(num_of_rows,num_of_columns,traffic_file,arch_file);

  printf("\nFinal free packet delay=%f,\nOriginal free packet delay:%f,\nImprovement:%f\% \n"
         ,curr_rate,ori_tao, (ori_tao-curr_rate)/ori_tao*100);  
  
  cout <<"\nThe architecture and routing table are written to " << arch_file <<endl;  
 cout <<"\n**********************************************************\n\n";

} // int main(int argn, char **argv)

void print_help() {

  cout << "Example Usage: " << endl;
  cout << " ./SmallNoC -num_of_rows 4 -num_of_columns 4 -constraint 10 -traffic_file traffic.config -architecture_file output.config \n";
  
  cout << "List of command line parameters:\n";
  cout << "-h:" << endl;
  cout << "\t-h or no arguments prints out this help." << endl <<endl;

  cout << "-v:" << endl;
  cout << "\trun in verbose mode." << endl << endl;

  cout << "-num_of_rows int:" << endl;
  cout << "\tsetup the number of rows of the underlying mesh network\n"
       << "\t(default = 4)\n\n";

  cout << "-num_of_columns int:" << endl;
  cout << "\tsetup the number of columns of the underlying mesh network\n"
       << "\t(default = 4)\n\n";

  cout << "-constraint int:" << endl;
  cout << "\tThe wire length constraint in terms of number of regular \n"
       << "\tlink segments (default = 10)\n\n";

  cout << "-traffic_file string:" << endl;
  cout << "\tThe configuration file that describes the application\n\n";

  cout << "-architecture_file string:" << endl;
  cout << "\tThe configuration file that describes the output topology and\n"
       <<"\trouting algorithm\n\n";
 
}

bool parse_parameters(int argn, char ** argv,int &num_of_rows, int &num_of_columns, int &constraint, char **arch_file,char **traffic_file, bool &verbose) {
 
  for (int i=1; i<argn; i++) {
      
    if (strcmp(argv[i], "-h") == 0) {
      print_help();
      return 0;        
    }
    
    if (strcmp(argv[i], "-v") == 0) {
      verbose = true;
      continue;
    }
    
    if (strcmp(argv[i], "-num_of_rows") == 0) {
      if ( (i+1==argn) || (!sscanf(argv[++i], "%d", &num_of_rows)) ) {
        cerr <<"Error reading in the number of rows!\n";
        return 0;
      }
      continue;
    }

    if (strcmp(argv[i], "-num_of_columns") == 0) {
      if  ( (i+1==argn) || (!sscanf(argv[++i], "%d", &num_of_columns)) ) {
        cerr <<"Cannot read the number of columns!\n";
        return 0;
      }
      continue;
    }
    
    if (strcmp(argv[i], "-constraint") == 0) {
      if  ( (i+1==argn) || (!sscanf(argv[++i], "%d", &constraint)) ) {
        cerr <<"Cannot read the link length constraint!\n";
        return 0;
      }
      continue;
    }
    
    if (strcmp(argv[i], "-traffic_file") == 0) {
      if  ( (i+1==argn) || ( !(*traffic_file = argv[++i]) ) ) {
        cerr <<"Cannot read the traffic file!\n";
        return 0;
      }
      continue;
    }
    
    if (strcmp(argv[i], "-architecture_file") == 0) {
      if  ( (i+1==argn) || ( !(*arch_file = argv[++i]) ) ) {
        cerr <<"Cannot read the architecture file!\n";
        return 0;
      }
      continue;
    }    

    cerr <<"Unrecognized option: " << argv[i] <<endl;
    return 0;
    
  } // for (int i=1; i<argn; i++)

}
